@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import collections
 import contextlib
 import curses
@@ -12,15 +14,11 @@ from typing import Callable
 from typing import cast
 from typing import Generator
 from typing import IO
-from typing import List
 from typing import Match
 from typing import NamedTuple
-from typing import Optional
 from typing import Pattern
-from typing import Tuple
 from typing import TYPE_CHECKING
 from typing import TypeVar
-from typing import Union
 
 from babi.buf import Buf
 from babi.buf import Modification
@@ -42,7 +40,7 @@ TCallable = TypeVar('TCallable', bound=Callable[..., Any])
 WS_RE = re.compile(r'^\s*')
 
 
-def get_lines(sio: IO[str]) -> Tuple[List[str], str, bool, str]:
+def get_lines(sio: IO[str]) -> tuple[list[str], str, bool, str]:
     sha256 = hashlib.sha256()
     lines = []
     newlines = collections.Counter({'\n': 0})  # default to `\n`
@@ -62,9 +60,24 @@ def get_lines(sio: IO[str]) -> Tuple[List[str], str, bool, str]:
     return lines, nl, mixed, sha256.hexdigest()
 
 
+class OpenError(RuntimeError):
+    pass
+
+
+def _load_file(filename: str) -> tuple[list[str], str, bool, str]:
+    try:
+        with open(filename, encoding='UTF-8', newline='') as f:
+            return get_lines(f)
+    except UnicodeDecodeError:
+        raise OpenError(f'error! not utf-8: {filename!r}')
+    except OSError:
+        # XXX: not quite correct, but maybe fix another day
+        raise OpenError(f'error! not a file: {filename!r}')
+
+
 class Action:
     def __init__(
-            self, *, name: str, modifications: List[Modification],
+            self, *, name: str, modifications: list[Modification],
             start_x: int, start_y: int, start_modified: bool,
             end_x: int, end_y: int, end_modified: bool,
             final: bool,
@@ -79,7 +92,7 @@ class Action:
         self.end_modified = end_modified
         self.final = final
 
-    def apply(self, file: 'File') -> 'Action':
+    def apply(self, file: File) -> Action:
         action = Action(
             name=self.name, modifications=file.buf.apply(self.modifications),
             start_x=self.end_x, start_y=self.end_y,
@@ -98,7 +111,7 @@ class Action:
 
 def action(func: TCallable) -> TCallable:
     @functools.wraps(func)
-    def action_inner(self: 'File', *args: Any, **kwargs: Any) -> Any:
+    def action_inner(self: File, *args: Any, **kwargs: Any) -> Any:
         self.finalize_previous_action()
         return func(self, *args, **kwargs)
     return cast(TCallable, action_inner)
@@ -111,7 +124,7 @@ def edit_action(
 ) -> Callable[[TCallable], TCallable]:
     def edit_action_decorator(func: TCallable) -> TCallable:
         @functools.wraps(func)
-        def edit_action_inner(self: 'File', *args: Any, **kwargs: Any) -> Any:
+        def edit_action_inner(self: File, *args: Any, **kwargs: Any) -> Any:
             with self.edit_action_context(name, final=final):
                 return func(self, *args, **kwargs)
         return cast(TCallable, edit_action_inner)
@@ -120,7 +133,7 @@ def edit_action(
 
 def keep_selection(func: TCallable) -> TCallable:
     @functools.wraps(func)
-    def keep_selection_inner(self: 'File', *args: Any, **kwargs: Any) -> Any:
+    def keep_selection_inner(self: File, *args: Any, **kwargs: Any) -> Any:
         with self.select():
             return func(self, *args, **kwargs)
     return cast(TCallable, keep_selection_inner)
@@ -128,7 +141,7 @@ def keep_selection(func: TCallable) -> TCallable:
 
 def clear_selection(func: TCallable) -> TCallable:
     @functools.wraps(func)
-    def clear_selection_inner(self: 'File', *args: Any, **kwargs: Any) -> Any:
+    def clear_selection_inner(self: File, *args: Any, **kwargs: Any) -> Any:
         ret = func(self, *args, **kwargs)
         self.selection.clear()
         return ret
@@ -143,7 +156,7 @@ class Found(NamedTuple):
 class _SearchIter:
     def __init__(
             self,
-            file: 'File',
+            file: File,
             reg: Pattern[str],
             *,
             offset: int,
@@ -155,7 +168,7 @@ class _SearchIter:
         self._start_x = file.buf.x + offset
         self._start_y = file.buf.y
 
-    def __iter__(self) -> '_SearchIter':
+    def __iter__(self) -> _SearchIter:
         return self
 
     def _stop_if_past_original(self, y: int, match: Match[str]) -> Found:
@@ -168,7 +181,7 @@ class _SearchIter:
             raise StopIteration()
         return Found(y, match)
 
-    def __next__(self) -> Tuple[int, Match[str]]:
+    def __next__(self) -> tuple[int, Match[str]]:
         x = self.file.buf.x + self.offset
         y = self.file.buf.y
 
@@ -200,25 +213,28 @@ class _SearchIter:
 class File:
     def __init__(
             self,
-            filename: Optional[str],
+            filename: str | None,
             initial_line: int,
             color_manager: ColorManager,
-            hl_factories: Tuple[HLFactory, ...],
+            hl_factories: tuple[HLFactory, ...],
+            *,
+            is_stdin: bool,
     ) -> None:
         self.filename = filename
         self.initial_line = initial_line
+        self.is_stdin = is_stdin
         self.modified = False
         self.buf = Buf([])
         self.nl = '\n'
-        self.sha256: Optional[str] = None
+        self.sha256: str | None = None
         self._in_edit_action = False
-        self.undo_stack: List[Action] = []
-        self.redo_stack: List[Action] = []
+        self.undo_stack: list[Action] = []
+        self.redo_stack: list[Action] = []
         self._hl_factories = hl_factories
         self._trailing_whitespace = TrailingWhitespace(color_manager)
         self._replace_hl = Replace()
         self.selection = Selection()
-        self._file_hls: Tuple[FileHL, ...] = ()
+        self._file_hls: tuple[FileHL, ...] = ()
 
     def ensure_loaded(
             self,
@@ -229,22 +245,23 @@ class File:
         if self.buf:
             return
 
-        if self.filename == '-':
+        if self.is_stdin:
             status.update('(from stdin)')
+            self.is_stdin = False
             self.filename = None
             self.modified = True
             sio = io.StringIO(stdin)
             lines, self.nl, mixed, self.sha256 = get_lines(sio)
-        elif self.filename is not None and os.path.isfile(self.filename):
-            with open(self.filename, encoding='UTF-8', newline='') as f:
-                lines, self.nl, mixed, self.sha256 = get_lines(f)
+        elif self.filename is not None and os.path.lexists(self.filename):
+            try:
+                lines, self.nl, mixed, self.sha256 = _load_file(self.filename)
+            except OpenError as e:
+                status.update(str(e))
+                self.filename = None
+                lines, self.nl, mixed, self.sha256 = get_lines(io.StringIO(''))
         else:
             if self.filename is not None:
-                if os.path.lexists(self.filename):
-                    status.update(f'{self.filename!r} is not a file')
-                    self.filename = None
-                else:
-                    status.update('(new file)')
+                status.update('(new file)')
             lines, self.nl, mixed, self.sha256 = get_lines(io.StringIO(''))
 
         self.buf = Buf(lines, self.buf.tab_size)
@@ -272,6 +289,14 @@ class File:
     def __repr__(self) -> str:
         return f'<{type(self).__name__} {self.filename!r}>'
 
+    def reset_modified_state(self) -> None:
+        for stack in (self.undo_stack, self.redo_stack):
+            first = True
+            for action in reversed(stack):
+                action.end_modified = not first
+                action.start_modified = True
+                first = False
+
     # movement
 
     @action
@@ -395,14 +420,14 @@ class File:
     @clear_selection
     def replace(
             self,
-            screen: 'Screen',
+            screen: Screen,
             reg: Pattern[str],
             replace: str,
     ) -> None:
         self.finalize_previous_action()
 
         count = 0
-        res: Union[str, PromptResult] = ''
+        res: str | PromptResult = ''
         search = _SearchIter(self, reg, offset=0)
         for line_y, match in search:
             end = match.end()
@@ -467,6 +492,45 @@ class File:
             self.buf.y = self.buf.file_y = pos
         self.buf.x = 0
 
+    @action
+    def alt_up(self, margin: Margin) -> None:
+        if self.buf.y > 0:
+            offset = 1
+            while self.buf[self.buf.y - offset] == '':
+                offset += 1
+            if offset >= 2:
+                self.buf.y -= offset
+                self.buf.scroll_screen_if_needed(margin)
+                self.buf.x = 0
+                return
+            while (
+                self.buf[self.buf.y - offset - 1] != '' and
+                self.buf.y - offset - 1 >= 0
+            ):
+                offset += 1
+            self.buf.y -= offset
+            self.buf.scroll_screen_if_needed(margin)
+            self.buf.x = 0
+
+    @action
+    def alt_down(self, margin: Margin) -> None:
+        if self.buf.y < len(self.buf) - 1:
+            offset = 0
+            while self.buf[self.buf.y + offset] != '':
+                offset += 1
+            if offset > 1:
+                self.buf.y += offset - 1
+                self.buf.scroll_screen_if_needed(margin)
+                self.buf.x = 0
+                return
+            while self.buf[self.buf.y + offset] == '':
+                if self.buf.y + offset >= len(self.buf) - 1:
+                    break
+                offset += 1
+            self.buf.y += offset
+            self.buf.scroll_screen_if_needed(margin)
+            self.buf.x = 0
+
     # editing
 
     @edit_action('backspace text', final=False)
@@ -597,7 +661,7 @@ class File:
 
     @edit_action('cut selection', final=True)
     @clear_selection
-    def cut_selection(self, margin: Margin) -> Tuple[str, ...]:
+    def cut_selection(self, margin: Margin) -> tuple[str, ...]:
         ret = []
         (s_y, s_x), (e_y, e_x) = self.selection.get()
         if s_y == e_y:
@@ -617,7 +681,7 @@ class File:
         self.buf.scroll_screen_if_needed(margin)
         return tuple(ret)
 
-    def cut(self, cut_buffer: Tuple[str, ...]) -> Tuple[str, ...]:
+    def cut(self, cut_buffer: tuple[str, ...]) -> tuple[str, ...]:
         # only continue a cut if the last action is a non-final cut
         if not self._continue_last_action('cut'):
             cut_buffer = ()
@@ -630,7 +694,7 @@ class File:
                 self.buf.x = 0
                 return cut_buffer + (victim,)
 
-    def _uncut(self, cut_buffer: Tuple[str, ...], margin: Margin) -> None:
+    def _uncut(self, cut_buffer: tuple[str, ...], margin: Margin) -> None:
         for cut_line in cut_buffer:
             line = self.buf[self.buf.y]
             before, after = line[:self.buf.x], line[self.buf.x:]
@@ -641,14 +705,14 @@ class File:
 
     @edit_action('uncut', final=True)
     @clear_selection
-    def uncut(self, cut_buffer: Tuple[str, ...], margin: Margin) -> None:
+    def uncut(self, cut_buffer: tuple[str, ...], margin: Margin) -> None:
         self._uncut(cut_buffer, margin)
 
     @edit_action('uncut selection', final=True)
     @clear_selection
     def uncut_selection(
             self,
-            cut_buffer: Tuple[str, ...], margin: Margin,
+            cut_buffer: tuple[str, ...], margin: Margin,
     ) -> None:
         self._uncut(cut_buffer, margin)
         self.buf.up(margin)
@@ -666,7 +730,7 @@ class File:
         self.buf.x = 0
         self.buf.scroll_screen_if_needed(margin)
 
-    def _selection_lines(self) -> Tuple[int, int]:
+    def _selection_lines(self) -> tuple[int, int]:
         (s_y, _), (e_y, _) = self.selection.get()
         e_y = min(e_y + 1, len(self.buf) - 1)
         if self.buf[e_y - 1] == '':
@@ -739,6 +803,30 @@ class File:
             else:
                 self._comment_add(lineno, prefix, minimum_indent)
 
+    def reload(self, status: Status, margin: Margin) -> None:
+        assert self.filename is not None
+        try:
+            lines, nl, mixed, sha256 = _load_file(self.filename)
+        except OpenError as e:
+            status.update(f'reload: {e}')
+            return
+
+        self.selection.clear()
+        self.nl, self.sha256 = nl, sha256
+        with self.edit_action_context('reload', final=True):
+            self.buf.replace_lines(lines)
+
+        self.buf.fixup_position(margin)
+
+        if mixed:
+            status.update(
+                f'reloaded! (mixed newlines will be converted to {self.nl!r})',
+            )
+        else:
+            self.modified = False
+            self.reset_modified_state()
+            status.update('reloaded!')
+
     DISPATCH = {
         # movement
         b'KEY_UP': up,
@@ -759,6 +847,8 @@ class File:
         b'kLFT5': ctrl_left,
         b'kHOM5': ctrl_home,
         b'kEND5': ctrl_end,
+        b'kUP3': alt_up,
+        b'kDN3': alt_down,
         # editing
         b'KEY_BACKSPACE': backspace,
         b'KEY_DC': delete,
@@ -778,6 +868,8 @@ class File:
         b'kLFT6': keep_selection(ctrl_left),
         b'kHOM6': keep_selection(ctrl_home),
         b'kEND6': keep_selection(ctrl_end),
+        b'kUP4': keep_selection(alt_up),
+        b'kDN4': keep_selection(alt_down),
     }
 
     @edit_action('text', final=False)
@@ -852,12 +944,12 @@ class File:
 
     def move_cursor(
             self,
-            stdscr: 'curses._CursesWindow',
+            stdscr: curses._CursesWindow,
             margin: Margin,
     ) -> None:
         stdscr.move(*self.buf.cursor_position(margin))
 
-    def draw(self, stdscr: 'curses._CursesWindow', margin: Margin) -> None:
+    def draw(self, stdscr: curses._CursesWindow, margin: Margin) -> None:
         to_display = min(self.buf.displayable_count, margin.body_lines)
 
         for file_hl in self._file_hls:
