@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import contextlib
 import curses
 import enum
@@ -5,14 +7,12 @@ import hashlib
 import os
 import re
 import signal
+import sre_parse
 import sys
+from typing import Callable
 from typing import Generator
-from typing import List
 from typing import NamedTuple
-from typing import Optional
 from typing import Pattern
-from typing import Tuple
-from typing import Union
 
 from babi.color_manager import ColorManager
 from babi.file import Action
@@ -26,13 +26,12 @@ from babi.prompt import Prompt
 from babi.prompt import PromptResult
 from babi.status import Status
 
-if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
+if sys.version_info >= (3, 8):  # pragma: >=3.8 cover
     import importlib.metadata as importlib_metadata
-else:  # pragma: no cover (<py38)
+else:  # pragma: <3.8 cover
     import importlib_metadata
 
 VERSION_STR = f'babi v{importlib_metadata.version("babi")}'
-EditResult = enum.Enum('EditResult', 'EXIT NEXT PREV OPEN')
 
 # TODO: find a place to populate these, surely there's a database somewhere
 SEQUENCE_KEYNAME = {
@@ -101,34 +100,59 @@ KEYNAME_REWRITE = {
 }
 
 
+class EditResult(enum.Enum):
+    EXIT = enum.auto()
+    EXIT_ALL = enum.auto()
+    EXIT_ALL_FORCE = enum.auto()
+    NEXT = enum.auto()
+    PREV = enum.auto()
+    OPEN = enum.auto()
+
+
+class Command(NamedTuple):
+    callback: Callable[[Screen, list[str]], EditResult | None]
+    nargs: str | int = 0
+
+
 class Key(NamedTuple):
-    wch: Union[int, str]
+    wch: int | str
     keyname: bytes
 
 
+class FileInfo(NamedTuple):
+    filename: str | None
+    initial_line: int
+    is_stdin: bool
+
+
 class Screen:
     def __init__(
             self,
-            stdscr: 'curses._CursesWindow',
-            filenames: List[Optional[str]],
-            initial_lines: List[int],
+            stdscr: curses._CursesWindow,
+            file_infos: list[FileInfo],
             perf: Perf,
     ) -> None:
         self.stdscr = stdscr
         self.color_manager = ColorManager.make()
         self.hl_factories = (Syntax.from_screen(stdscr, self.color_manager),)
         self.files = [
-            File(filename, line, self.color_manager, self.hl_factories)
-            for filename, line in zip(filenames, initial_lines)
+            File(
+                info.filename,
+                info.initial_line,
+                self.color_manager,
+                self.hl_factories,
+                is_stdin=info.is_stdin,
+            )
+            for info in file_infos
         ]
         self.i = 0
         self.history = History()
         self.perf = perf
         self.status = Status()
         self.margin = Margin.from_current_screen()
-        self.cut_buffer: Tuple[str, ...] = ()
+        self.cut_buffer: tuple[str, ...] = ()
         self.cut_selection = False
-        self._buffered_input: Union[int, str, None] = None
+        self._buffered_input: int | str | None = None
 
     @property
     def file(self) -> File:
@@ -145,7 +169,7 @@ class Screen:
             files = ''
             version_width = len(VERSION_STR) + 2
         centered = filename.center(self.margin.cols)[version_width:]
-        s = f' {VERSION_STR} {files}{centered}{files}'
+        s = f' {VERSION_STR} {files}{centered}'
         self.stdscr.insstr(0, 0, s, curses.A_REVERSE)
 
     def _get_sequence_home_end(self, wch: str) -> str:
@@ -271,8 +295,8 @@ class Screen:
     def quick_prompt(
             self,
             prompt: str,
-            opt_strs: Tuple[str, ...],
-    ) -> Union[str, PromptResult]:
+            opt_strs: tuple[str, ...],
+    ) -> str | PromptResult:
         opts = {opt[0] for opt in opt_strs}
         while True:
             x = 0
@@ -318,10 +342,10 @@ class Screen:
             prompt: str,
             *,
             allow_empty: bool = False,
-            history: Optional[str] = None,
+            history: str | None = None,
             default_prev: bool = False,
-            default: Optional[str] = None,
-    ) -> Union[str, PromptResult]:
+            default: str | None = None,
+    ) -> str | PromptResult:
         default = default or ''
         self.status.clear()
         if history is not None:
@@ -378,7 +402,7 @@ class Screen:
         else:
             self.file.uncut(self.cut_buffer, self.margin)
 
-    def _get_search_re(self, prompt: str) -> Union[Pattern[str], PromptResult]:
+    def _get_search_re(self, prompt: str) -> Pattern[str] | PromptResult:
         response = self.prompt(prompt, history='search', default_prev=True)
         if response is PromptResult.CANCELLED:
             return response
@@ -391,8 +415,8 @@ class Screen:
     def _undo_redo(
             self,
             op: str,
-            from_stack: List[Action],
-            to_stack: List[Action],
+            from_stack: list[Action],
+            to_stack: list[Action],
     ) -> None:
         if not from_stack:
             self.status.update(f'nothing to {op}!')
@@ -421,69 +445,131 @@ class Screen:
                 'replace with', history='replace', allow_empty=True,
             )
             if response is not PromptResult.CANCELLED:
-                self.file.replace(self, search_response, response)
+                try:
+                    sre_parse.parse_template(response, search_response)
+                except re.error:
+                    self.status.update('invalid replacement string')
+                else:
+                    self.file.replace(self, search_response, response)
 
-    def command(self) -> Optional[EditResult]:
+    def _command_w(self, args: list[str]) -> None:
+        self.save()
+
+    def _command_wq(self, args: list[str]) -> EditResult:
+        self.save()
+        return EditResult.EXIT
+
+    def _command_sort(self, args: list[str]) -> None:
+        if self.file.selection.start:
+            self.file.sort_selection(self.margin)
+        else:
+            self.file.sort(self.margin)
+        self.status.update('sorted!')
+
+    def _command_sort_bang(self, args: list[str]) -> None:
+        if self.file.selection.start:
+            self.file.sort_selection(self.margin, reverse=True)
+        else:
+            self.file.sort(self.margin, reverse=True)
+        self.status.update('sorted!')
+
+    def _command_tabsize(self, args: list[str]) -> None:
+        tab_size, = args
+        try:
+            parsed_tab_size = int(tab_size)
+        except ValueError:
+            self.status.update(f'invalid size: {tab_size}')
+        else:
+            if parsed_tab_size <= 0:
+                self.status.update(f'invalid size: {parsed_tab_size}')
+            else:
+                for file in self.files:
+                    file.buf.set_tab_size(parsed_tab_size)
+                self.status.update('updated!')
+
+    def _command_expandtabs(self, args: list[str]) -> None:
+        for file in self.files:
+            file.buf.expandtabs = True
+        self.status.update('updated!')
+
+    def _command_noexpandtabs(self, args: list[str]) -> None:
+        for file in self.files:
+            file.buf.expandtabs = False
+        self.status.update('updated!')
+
+    def _command_comment(self, args: list[str]) -> None:
+        if len(args) == 1:
+            comment, = args
+        else:
+            comment = '#'
+        if self.file.selection.start:
+            self.file.toggle_comment_selection(comment)
+        else:
+            self.file.toggle_comment(comment)
+
+    def _command_reload(self, args: list[str]) -> None:
+        if self.file.filename is None:
+            self.status.update('file has not been saved yet!')
+            return
+
+        if self.file.modified:
+            response = self.quick_prompt(
+                'reload will discard changes - continue', ('yes', 'no'),
+            )
+            if response != 'y':
+                return
+
+        self.file.reload(self.status, self.margin)
+
+    COMMANDS = {
+        ':qall': Command(lambda self, args: EditResult.EXIT_ALL),
+        ':qall!': Command(lambda self, args: EditResult.EXIT_ALL_FORCE),
+        ':q': Command(lambda self, args: self.quit_save_modified()),
+        ':q!': Command(lambda self, args: EditResult.EXIT),
+        ':w': Command(_command_w),
+        ':wq': Command(_command_wq),
+        ':sort': Command(_command_sort),
+        ':sort!': Command(_command_sort_bang),
+        ':tabstop': Command(_command_tabsize, nargs=1),
+        ':tabsize': Command(_command_tabsize, nargs=1),
+        ':expandtabs': Command(_command_expandtabs),
+        ':noexpandtabs': Command(_command_noexpandtabs),
+        ':comment': Command(_command_comment, nargs='?'),
+        ':reload': Command(_command_reload),
+    }
+
+    def command(self) -> EditResult | None:
         response = self.prompt('', history='command')
         if response is PromptResult.CANCELLED:
-            pass
-        elif response == ':q':
-            return self.quit_save_modified()
-        elif response == ':q!':
-            return EditResult.EXIT
-        elif response == ':w':
-            self.save()
-        elif response == ':wq':
-            self.save()
-            return EditResult.EXIT
-        elif response == ':sort':
-            if self.file.selection.start:
-                self.file.sort_selection(self.margin)
-            else:
-                self.file.sort(self.margin)
-            self.status.update('sorted!')
-        elif response == ':sort!':
-            if self.file.selection.start:
-                self.file.sort_selection(self.margin, reverse=True)
-            else:
-                self.file.sort(self.margin, reverse=True)
-            self.status.update('sorted!')
-        elif response.startswith((':tabstop ', ':tabsize ')):
-            _, _, tab_size = response.partition(' ')
-            try:
-                parsed_tab_size = int(tab_size)
-            except ValueError:
-                self.status.update(f'invalid size: {tab_size}')
-            else:
-                if parsed_tab_size <= 0:
-                    self.status.update(f'invalid size: {parsed_tab_size}')
-                else:
-                    for file in self.files:
-                        file.buf.set_tab_size(parsed_tab_size)
-                    self.status.update('updated!')
-        elif response.startswith(':expandtabs'):
-            for file in self.files:
-                file.buf.expandtabs = True
-            self.status.update('updated!')
-        elif response.startswith(':noexpandtabs'):
-            for file in self.files:
-                file.buf.expandtabs = False
-            self.status.update('updated!')
-        elif response == ':comment' or response.startswith(':comment '):
-            _, _, comment = response.partition(' ')
-            comment = (comment or '#').strip()
-            if self.file.selection.start:
-                self.file.toggle_comment_selection(comment)
-            else:
-                self.file.toggle_comment(comment)
-        else:
-            self.status.update(f'invalid command: {response}')
-        return None
+            return None
 
-    def save(self) -> Optional[PromptResult]:
+        cmd, *args = response.split()
+        if cmd not in self.COMMANDS:
+            self.status.update(f'invalid command: {response}')
+            return None
+
+        command = self.COMMANDS[cmd]
+        if isinstance(command.nargs, int):
+            if len(args) != command.nargs:
+                self.status.update(
+                    f'`{cmd}`: expected {command.nargs} args '
+                    f'but got {len(args)}',
+                )
+                return None
+        elif command.nargs == '?':
+            if len(args) not in {0, 1}:
+                self.status.update(
+                    f'`{cmd}`: expected 0 or 1 args but got {len(args)}',
+                )
+                return None
+        else:
+            raise NotImplementedError(command)
+
+        return command.callback(self, args)
+
+    def save(self) -> PromptResult | None:
         self.file.finalize_previous_action()
 
-        # TODO: make directories if they don't exist
         # TODO: maybe use mtime / stat as a shortcut for hashing below
         # TODO: strip trailing whitespace?
         # TODO: save atomically?
@@ -494,21 +580,23 @@ class Screen:
             else:
                 self.file.filename = filename
 
-        if os.path.isfile(self.file.filename):
+        if not os.path.isfile(self.file.filename):
+            sha256: str | None = None
+        else:
             with open(self.file.filename, encoding='UTF-8', newline='') as f:
                 *_, sha256 = get_lines(f)
-        else:
-            sha256 = hashlib.sha256(b'').hexdigest()
 
         contents = self.file.nl.join(self.file.buf)
         sha256_to_save = hashlib.sha256(contents.encode()).hexdigest()
 
         # the file on disk is the same as when we opened it
-        if sha256 not in (self.file.sha256, sha256_to_save):
+        if sha256 not in (None, self.file.sha256, sha256_to_save):
             self.status.update('(file changed on disk, not implemented)')
             return PromptResult.CANCELLED
 
         try:
+            dir_path = os.path.dirname(os.path.abspath(self.file.filename))
+            os.makedirs(dir_path, exist_ok=True)
             with open(
                 self.file.filename, 'w', encoding='UTF-8', newline='',
             ) as f:
@@ -522,17 +610,10 @@ class Screen:
         num_lines = len(self.file.buf) - 1
         lines = 'lines' if num_lines != 1 else 'line'
         self.status.update(f'saved! ({num_lines} {lines} written)')
-
-        # fix up modified state in undo / redo stacks
-        for stack in (self.file.undo_stack, self.file.redo_stack):
-            first = True
-            for action in reversed(stack):
-                action.end_modified = not first
-                action.start_modified = True
-                first = False
+        self.file.reset_modified_state()
         return None
 
-    def save_filename(self) -> Optional[PromptResult]:
+    def save_filename(self) -> PromptResult | None:
         response = self.prompt('enter filename', default=self.file.filename)
         if response is PromptResult.CANCELLED:
             return PromptResult.CANCELLED
@@ -540,16 +621,22 @@ class Screen:
             self.file.filename = response
             return self.save()
 
-    def open_file(self) -> Optional[EditResult]:
+    def open_file(self) -> EditResult | None:
         response = self.prompt('enter filename', history='open')
         if response is not PromptResult.CANCELLED:
-            opened = File(response, 0, self.color_manager, self.hl_factories)
+            opened = File(
+                response,
+                0,
+                self.color_manager,
+                self.hl_factories,
+                is_stdin=False,
+            )
             self.files.append(opened)
             return EditResult.OPEN
         else:
             return None
 
-    def quit_save_modified(self) -> Optional[EditResult]:
+    def quit_save_modified(self) -> EditResult | None:
         if self.file.modified:
             response = self.quick_prompt(
                 'file is modified - save', ('yes', 'no'),
@@ -597,14 +684,14 @@ class Screen:
     }
 
 
-def _init_screen() -> 'curses._CursesWindow':
+def _init_screen() -> curses._CursesWindow:
     # set the escape delay so curses does not pause waiting for sequences
     if (
             sys.version_info >= (3, 9) and
             hasattr(curses, 'set_escdelay')
-    ):  # pragma: no cover
+    ):  # pragma: >=3.9 cover
         curses.set_escdelay(25)
-    else:  # pragma: no cover
+    else:  # pragma: <3.9 cover
         os.environ.setdefault('ESCDELAY', '25')
 
     stdscr = curses.initscr()
@@ -623,7 +710,7 @@ def _init_screen() -> 'curses._CursesWindow':
 
 
 @contextlib.contextmanager
-def make_stdscr() -> Generator['curses._CursesWindow', None, None]:
+def make_stdscr() -> Generator[curses._CursesWindow, None, None]:
     """essentially `curses.wrapper` but split out to implement ^Z"""
     try:
         yield _init_screen()
